#ifndef COMDefH
#define COMDefH

// Class TUnknown is the basic implementation of an object derived from
// the COM abstract interface class IUnknown, defined in windows.h or in
// module COMBase.h for applications that don't include windows.h.
//
// To avoid using multiple inheritance (allows translation of these modules
// into Delphi Object Pascal, which only allows single inheritance), the
// inheritance tree will be
//
//     IUnknown        // Root interface for all COM objects
//     TUnknown        // Implements IUnknown QueryInterface(), 
//                     // AddRef(), Release()
//     IMyCOMInterface // Extends IUnknown with virtual abstract methods 
//     TMyCOMObject    // Implementation of IMyCOMInterface.
//
// A DLL module will have two situations -- a case where it just needs an
// interface to an object in another module, and a case where it needs to
// create an object with that interface.  In the first case, where you only
// need to use a particular COM object but not create it, class TUnknown may
// be left out of the inheritance tree.  In the second case, where you need
// to create an object with the IMyCOMInterface or other interface, TUnknown
// needs to be in the inheritance tree.  Since TUnknown does not implement any
// additional virtual functions, it does not disrupt the extension of
// IUnknown by IMyCOMInterface by adding any virtual functions that are not
// part of the specified interface.
//
// One "glitch" in C++ not found in Delphi Object Pascal is that a virtual
// destructor in C++ occupies a slot in the virtual method table (v-table).
// If you define a virtual destructor, that will throw off the ordering of
// interface v-table and generate runtime errors when using that interface
// across module boundaries.  Release() triggers object deletion; note that
// there is no virtual destructor to override to perform actions when Release()
// deletes the object.  If you need the behavior of a virtual destructor, you
// must override and reimplement Release() -- you can call the non-virtual
// decRefCount() to help with this.
//
// Microsoft has its own way of creating COM objects in C++ using 1) inheritance
// 2) templates, 3) macros, and 3) cut-and-paste in the form of code-generation 
// "Wizards" -- this module supports a pure inheritance-based approach.
//
// The implementation of QueryInterface in TUnknown acknowledges implementation
// of interface IUnknown as specified by the 128-bit GUID assigned to IUnknown
// by Microsoft.  For COM modules used in a restrictive setting, an IUnknown
// interface can be simply cast to the interface that an object supports without
// going through QueryInterface() -- this way derived interfaces don't need to
// be assigned GUIDs other than IUnknown.  In situations where the correct 
// interface must be supplied by QueryInterface(), you will have to obtain a 
// GUID for your interface using available tools for generating one, and you 
// will have to override QueryInterface() to supply the correct response to that 
// GUID along with calling the base-class QueryInterface() to get the right 
// response to requests for IUnknown. The non-virtual (doesn't show up in the 
// interface v-table) helper function IsEqualGUID() can assist with this task.

#ifdef _WIN32
#define STDCALL __stdcall
#define LIBEXPORT extern "C" __declspec(dllexport)
#else
#define STDCALL
#define LIBEXPORT extern "C"
#endif

class TUnknown : public IUnknown
{
private:
    int ref_count;

protected:
    ULONG decRefCount()
    {
        ref_count--;
        if (ref_count <= 0)
            return 0;
        else
            return ref_count;
    }

    bool IsEqualGUID(_GUID* id1, _GUID* id2)
    {
        return
            ((unsigned long *) id1)[0] == ((unsigned long *) id2)[0] &&
            ((unsigned long *) id1)[1] == ((unsigned long *) id2)[1] &&
            ((unsigned long *) id1)[2] == ((unsigned long *) id2)[2] &&
            ((unsigned long *) id1)[3] == ((unsigned long *) id2)[3];
    }

public:
    virtual HRESULT STDCALL QueryInterface(const struct _GUID& guid, void** pvObj)
    {
        _GUID IID_IUnknown =
	    {0x00000000,0x0000,0x0000,{0xC0,0x00,0x00,0x00,0x00,0x00,0x00,0x46}};
        if (IsEqualGUID((_GUID*) &guid,(_GUID*) &IID_IUnknown))
        {
            AddRef();
            *pvObj = this;
            return S_OK;
        }
        else
        {
            *pvObj = NULL;
            return E_NOINTERFACE;
        }
    }

    virtual ULONG STDCALL AddRef()
    {
        ref_count++;
        return ref_count;
    }

    virtual ULONG STDCALL Release()
    {
        int rc = decRefCount();
        if (rc <= 0) delete this;
        return rc;
    }

    TUnknown() : ref_count(0) {}
};

#endif
