// Copyright 2007 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

// NonCausalP parallel filter -- requires the NonCausal base class
// as revised for parallel-filter composition.
//
// Combines this and the parallel element filtP -- if filtP is in turn a 
// NonCausalP, it too can have a parallel element, allowing for an 
// arbitrary number of parallel elements.
//
// The default Apply(src,dst,nsamp) method of base class NonCausalP sets 
// the dst array.  The Apply(src,dst,nsamp) method of filtP can combine
// its output with the samples already in the dst array from this, the
// previous parallel stage.

#ifndef NonCausalPH
#define NonCausalPH

#include "NonCausal.h"

class NonCausalP : public NonCausal
{
private:
	NonCausal* filtP;

public:
	NonCausalP() : filtP(NULL) {}

	NonCausalP(NonCausal* filtP) 
	{
		if (filtP != NULL)
			filtP->AddRef();
		this->filtP = filtP;
	}

	// Non-virtual destructor required -- Release() is virtual
    ~NonCausalP() 
    {
		if (filtP != NULL)
			filtP->Release();
    }

	// Classes implementing the IUnknown interface or its extensions must 
	// not have a virtual destructor because that produces an extraneous
	// entry in the virtual method table (v-table).  You must override and
	// reimplement Release(), which has a v-table slot reserved by IUnknown, 
	// to invoke the non-virtual destructor of this, the derived class.
    virtual ULONG STDCALL Release() 
    {
        int rc = decRefCount();
        if (rc <= 0) delete this;
        return rc;
    }

	virtual bool STDCALL EndOfSrc() 
	{
		if (filtP != NULL)
		    return NonCausal::EndOfSrc() && filtP->EndOfSrc();
		else
			return NonCausal::EndOfSrc();
	}
   
    virtual void STDCALL TransmitNow(float* wave_samp, int nBuf, int& nRet)
    {
		if (filtP == NULL)
			NonCausal::TransmitNow(wave_samp,nBuf,nRet);
		else 
		{
			// Limit nBuf to capacity of this and filtP
			int nBuf0;
		    NonCausal::TransmitNow(NULL,nBuf,nBuf0);
		    filtP->TransmitNow(NULL,nBuf,nBuf);
            if (nBuf > nBuf0)
			    nBuf = nBuf0;
			
			// Filter this followed by filter filtP
			NonCausal::TransmitNow(wave_samp,nBuf,nRet);
		    filtP->TransmitNow(wave_samp,nBuf,nRet);
		}
    }

    virtual void  STDCALL Set_src_list(IWaveSourceC** src_list, int nSrc) 
    {
		if (filtP == NULL)
			// Takes only first element -- releases the rest
			NonCausal::Set_src_list(src_list,nSrc);
		else
		{
            // Pass only first element
			NonCausal::Set_src_list(src_list,1);
			// Takes what it needs from [1 .. nSrc-1], releases the rest
			filtP->Set_src_list(src_list+1,nSrc-1);
		}
    }

    virtual void  STDCALL Set_console_out(IConsoleOutC* console_out)
    {
		NonCausal::Set_console_out(console_out);
		if (filtP != NULL)
		    filtP->Set_console_out(console_out);
    }

	virtual void  STDCALL Set_param_list(double* param_list, int nParam) 
	{
		if (filtP != NULL)
		    filtP->Set_param_list(param_list,nParam);
	}
};
#endif
